@@ -364,7 +364,7 @@ class Dataset(object):
     def _set_properties(self, api_response):
         """Update properties from resource in body of ``api_response``
 
-        :type api_response: httplib2.Response
+        :type api_response: dict
         :param api_response: response returned from an API call.
         """
         self._properties.clear()
